import numpy as np
from math import pi, sin, cos
import cmath
from typing import Union, List
from numpy.core.numeric import full

def im_to_tuple(im):
    return (im.real, im.imag)

def tuple_to_im(t):
    return complex(t[0],t[1])

def mat_im_to_tuple(M_i):
    dim = np.shape(M_i)
    M_t = np.zeros(dim).tolist()
    if len(dim) == 1:
        for i in range(dim[0]):
            M_t[i] = im_to_tuple(M_i[i])
    elif len(dim) == 2:
        for i in range(dim[0]):
            for j in range(dim[1]):
                M_t[i][j] = im_to_tuple(M_i[i][j])
    else:
        raise ValueError("the imput is not a vector or a matrix")
    return M_t
    # this is a list of list, not a numpy array

def mat_tuple_to_im(M_t):
    dim = np.shape(M_t)
    print(dim)
    M_i = np.zeros(dim).tolist()
    if len(dim) == 2:
        for i in range(dim[0]):
            M_i[i] = tuple_to_im(M_t[i])
    elif len(dim) == 3:
        for i in range(dim[0]):
            for j in range(dim[1]):
                M_i[i][j] = tuple_to_im(M_t[i][j])
    else:
        raise ValueError("the imput is not a vector or a matrix")
    return np.array(M_i, dtype=np.complex128)



def float_to_bin(x, n_digits:int):
    """ 
    Convert a number x in range [0,1] to a binary string truncated to length n_digits

    arguments:
        x: float 
        n_digits: integer
    
    return: 
        x_bin: string
        The decimal representation of digits AFTER '0.'
        Ex:
            Input 0.75 has binary representation 0.11 
            Then this function would return '11'
    """
    if x < 0 or x >= 1:
        raise ValueError("x must be in interval [0,1)")
    x_round = round(x * 2**n_digits)
    # print(x_round)
    # print(2**n_digits)
    if x_round == 2**n_digits:
        x_round = 0
    x_raw = bin(x_round)
    x_bin = x_raw[2:].zfill(n_digits)
    return x_bin

def bin_to_float(b:Union[str, list]):
    if isinstance(b, list):
        s = ""
        # print(b)
        for digit in b:
            if len(digit) != 1:
                raise ValueError("Invalid input")
            s += digit
    if b[:2] == "0.":
        b = b[2:]
    f = 0
    for d_i, digit in enumerate(b, 1):
        if digit not in "01":
            # print(digit)
            raise ValueError("Invalid input")
        if digit == "1":
            f += 2**-d_i

    return f

def phase_to_exp(phi, radians = False):
    """
    Returns the complex exponential e^(i*phi*2*pi)
    If radians == True it returns e^(i*phi) 
    """
    phi = np.array(phi)
    if radians:
        phase = phi*1j
    else:
        phase = phi*2*pi*1j
    return np.exp(phase)

def count_conversion(counts):
    """Converts a dictionary of combined measurement counts to a list with the number of individual counts per classical channel
    Args:
        counts: dict
            The dictionary version of the qiskit.result.counts object
    
    returns:
        ones: 1D numpy array
            The count of the individual bits, in particular the number of ones
            first element = number of times a 1 was measured in the first register
    
    Ex:
        input: counts = {'00': 5, "01": 10, "10": 100, "11": 500}
        return: [600, 510]
    """
    if not isinstance(counts, dict):
        try:
            counts = dict(counts)
        except:
            raise TypeError("Input counts is not dict or convertable to dict")
    
    n_bits = len(list(counts.keys())[0])
    ones = np.zeros(n_bits)
    for bit_string in counts:
        result_count = counts[bit_string]
        for d_i, digit in enumerate(bit_string):
            if digit == "1":
                ones[d_i] += result_count
    return ones

def get_sub_bitstring_counter(counts, n_splits):
    if not isinstance(counts, dict):
        try:
            counts = dict(counts)
        except:
            raise TypeError("Input counts is not dict or convertable to dict")

    n_bits = len(list(counts.keys())[0])
    if n_bits%n_splits != 0:
        raise ValueError("Length of bitstring and n_splits not compatible")
    sub_bitstring_len = n_bits//n_splits
    sub_bitstring_dicts = []
    for n in range(n_splits):
        d = {}
        b_start = n*sub_bitstring_len
        b_end = (n+1)*sub_bitstring_len
        for full_bitstring in counts:
            temp_count = counts[full_bitstring]
            sub_bitstring = full_bitstring[b_start:b_end]
            if sub_bitstring in d.keys():
                d[sub_bitstring] += temp_count
            else:
                d[sub_bitstring] = temp_count
        sub_bitstring_dicts.append(d)
    
    return sub_bitstring_dicts

def calculate_bitstring_distribution(ones_list:list, bitstring_length:int, n_shots:int):
    
    # print()
    assert 2**bitstring_length -1 == len(ones_list)
    # indices = [2**j - 1 for j in range(bitstring_length)]
    # most_significant_digits = ones_list[-index:]
    bitstring_dict = {"":1}
    i_bit = 0
    while i_bit < bitstring_length:
        index = 2**i_bit
        ones = ones_list[-index:]
        new_bitstring_dict = {}
        for key in bitstring_dict:
            p_key = bitstring_dict[key]
            # print(p_key)
            key_index = 0 if key == "" else int(key, 2) 
            p_one = p_key * ones[key_index]/n_shots
            p_zero = p_key*(1 - ones[key_index]/n_shots)
            # print(p_one, p_zero)
            new_bitstring_dict["0"+key] = p_zero
            new_bitstring_dict["1" + key] = p_one
        bitstring_dict = new_bitstring_dict
        i_bit += 1
    return bitstring_dict

def normalize_counts(counts):
    norm_count = []
    for count in range(counts):
        if not isinstance(counts, dict):
            try:
                count = dict(count)
            except:
                raise TypeError("Input counts is not dict or convertable to dict")
    
        norm_count.append({key:count[key]/2 for key in count})

    return norm_count

if __name__ == "__main__":
    np.set_printoptions(precision=2)

    ones_list = [4,6,2, 34, 32, 9, 27]
    n_shots = 50
    b_len = 3

    k = calculate_bitstring_distribution(ones_list, b_len, n_shots)

    j = sum([k[l] for l in k])

    print(k)
    print(j)
    # k = np.arange(0,1,0.1)
    # # print(k)
    # l = phase_to_exp(k)
    # print(l)
    # test_dict = {'0000': 5, "0001": 10, "1101": 100, "1111": 500}
    # c = get_sub_bitstring_counter(test_dict, 2)
    # print(c)

    # n = 0.3
    # l = float_to_bin(n, 8)
    # print(l)
    # k = bin_to_float(l)
    # print(k)
    
    # h = "01001100"
    # print(bin_to_float(h))


    # M = np.array([[1+1j,0+0j],[1+1j,2+1j]], dtype=np.complex128)
    # M = np.zeros(2)
    # dim = np.shape(M)
    # print(len(dim))
    # M_new = mat_im_to_tuple(M)
    # print(M_new)
    # M_2 = mat_tuple_to_im(M_new)
    # print(M_2)


from QPE.Kitaev import Kitaev
from QPE.Iterative import Iterative
from QPE.QFT import QFT
from QPE.HelpFunctions import phase_to_exp
from QPE.Unitary import Unitary
import json
import numpy as np
import os
import re

class ExperimentSet():
    def __init__(self, instructions, draw_params:dict = {"NT":0, "T":0}):
        self.draw_params = draw_params 
        self.get_specs(instructions)
        self.PE_dicts = self.get_PE_dicts()
        self.generate_filenames()
        self.result_dicts = []
        # print(self.draw_params)
    
    def get_specs(self, instructions):
        self.path = instructions.pop("path")
        self.name = instructions.pop("name")
        self.filename_flags = instructions.pop("filename_flags")
        self.instructions = instructions

    def get_PE_dicts(self):
        PE_dicts = [{}]
        non_defaults = list(self.instructions.keys())

        def generate_combinations(PE_dicts, key, param_list):
            new_PE_dicts = []
            for old_dict in PE_dicts:
                for param in param_list:
                    new_d = old_dict.copy()
                    new_d[key] = param
                    new_PE_dicts.append(new_d)
            return new_PE_dicts

        while len(non_defaults) > 0:
            key = non_defaults.pop(-1)
            param_list = self.instructions[key]
            PE_dicts = generate_combinations(PE_dicts, key, param_list)
        
        print(f"Using {len(PE_dicts)} different combinations")
        return PE_dicts

    def generate_filenames(self):

        keys = list(self.PE_dicts[0].keys())
        estimator_pattern = r"[A-Za-z]+"
        estimator_pattern = r"QPE\.([A-Za-z]+)"

        # print(keys)
        # print(self.filename_flags)
        for d in self.PE_dicts:
            filename = ""
            for flag in self.filename_flags:
                if flag == "estimator":
                    full_e_string = str(d["estimator"])
                    match = re.search(estimator_pattern, full_e_string)

                    s = match.groups()[0]
                    filename += f"_{s}"
                elif flag == "U":
                    unitary_name = d["U"].name
                    if unitary_name != "":
                        filename += f"_{unitary_name}"
                elif flag == "backend_name":
                    filename += f"_{d['backend_params']['backend_name']}"
                elif flag == "backend_service":
                        filename += f"_{d['backend_params']['service']}"
                elif flag == "n_rotations":
                    filename += f"_n_rotations{d['method_specific_params']['n_rotations']}" 
                elif flag in ["n_digits","n_shots"]:
                    filename += f"_{flag}{d[flag]}"                   

            d["filename"] = filename



    def one_run(self, index:int, draw_params = {"T":-1, "NT":-1}):
        PE_dict = self.PE_dicts[index]
        filename = PE_dict.pop("filename")
        #try:
        estimator = PE_dict.pop("estimator")
        e = estimator(**PE_dict)
        e.run_circuit()
        e.post_process()
        # except:
        #     raise RuntimeError("Something went wrong with the PhaseEstimator")
        
        # print(self.draw_params)

        if (draw_params["T"] != 0) or (draw_params["NT"] != 0):
            plot_path = self.path + f"/plots/{filename}"
            if not os.path.isdir(plot_path):
                os.makedirs(plot_path, exist_ok=True)
                print(f"Created directory {plot_path}")
            
            drawing_path = plot_path + f"{filename}_{index}"
            e.draw_circuits(drawing_path, self.draw_params["NT"], self.draw_params["T"])

        dicts = e.get_dicts()
        json_path = self.path + "/data"

        if not os.path.isdir(json_path):
            os.makedirs(json_path, exist_ok=True)
            print(f"Created directory {json_path}")

        for d_i, d in enumerate(dicts):
            d["Experiment"] = self.name
            d["name"] = filename
            json_filepath = json_path + f"/{filename}-{index}_{d_i+1}.json"
            with open(json_filepath, "w") as write_file:
                json.dump(d, write_file, indent = True)
        
    def run(self):
        for i in range(len(self.PE_dicts)):
            # try:
            self.one_run(i, self.draw_params)
            # except:
            #     print(f"Experiment {self.name} failed for PhaseEstimator described by paramaters {self.PE_dicts[i]}")



if __name__ == "__main__":

    # estimators = [Kitaev, Iterative]
    diag_matrix = np.diag(phase_to_exp([0.3, 0.85]))
    unitaries = [
        Unitary(random=True, random_state=123, dim = 1),
        # Unitary(diag_matrix)
        ]
    n_digits_list = list(range(2,4))
    backend_param_list = [
        {
            "service": "local"
        }
        # ,
        # {
        #     "service": "IBMQ", 
        #     "filters": filters["5qubit"]
        # }
    ]
    state_dicts = [
        {"N_states": 2},
        {"eigen_coefs": [
            [1,1],
            [3,1]
        ]}
    ]

    instructions = {
    "name": "superexperiment",
    # "estimator": [Kitaev, Iterative],
    "estimator":[Iterative],
    "U": [Unitary(random=True, random_state=123, dim = 1)],
    "n_digits": list(range(2,4)),
    "backend_params": [{"service": "local"}],
    "n_shots": [500],
    "N_states": [2],
    "path": "Results/Experiments/superexperiment",
    "filename_flags": ["estimator", "backend_service", "n_digits"]
    }

    # path = "Results/Experiments/superexperiment"

    
    superexperiment = ExperimentSet(instructions)
    superexperiment.run()

# j = {
    
#     "Estimator": "Kitaev",
#     "Backend":{
#         "host": "local",
#         "simulator": True,
#         "nqubits": 5,
#         "name": "lima",
#     },

#     "Unitary": {
#         "dimension": 1,
#         "data": [
#             [(0.3, 0.6)],
#             []
#         ],
#         "eigenvectors":[
#             [(1,0),(0,0)],
#             [(1,0.5),(0.1,0.2)]
#         ],
#         "phis":[],
#         "phi_bitstrings": []
#     },

#     "Bits of precision": 3,
#     "Input state": [0,0.5,0.5,0],
    
#     "estimated phase": {
#         "phi": 0.3,
#         "bitstring": "0101",
#         "distribution": {
#             "00": 98,
#             "01": 309,
#             "10": 20,
#             "11": 764
#         }
#     }
# }


import os
import sys
sys.path.insert(0,os.getcwd())
from qiskit import IBMQ, QuantumCircuit, transpile, execute
from qiskit.providers.jobstatus import JobStatus
from QPE.Backend import GetBackend, Login
from QPE.Unitary import Unitary
from QPE.HelpFunctions import mat_im_to_tuple, mat_tuple_to_im
import numpy as np
import time
from typing import List, Union
from datetime import datetime
import json

class PhaseEstimator():

    """
    Superclass to algorithms for Quantum Phase Estimation

    Only one of the arguments 'N_states', 'input_states' or 'eigen_coefs' is needed

    Parameters
    ----------
    U: Unitary
        The unitary to perform phase estimation on
    
    n_digits: int
        The number of bits of precision the algorithm should use
    
    N_states: int
        The number of eigenstates to estimate the phase of. If parameter is set to -1, all states are estimated
    
    input_states: List[list]
        The input states that is initialized and estimated, in the computational basis. Each sublist should contain the complex coefficients that describe the desired statevector.
    
    eigen_coefs: List[list]
        The input states that is initialized and estimated, but in the eigenvector basis of the unitary. Each sublist should contain the complex coefficients that describe the desired statevector in the computational basis.

    n_shots: int
        Number of shots that we run the associated QuantumCircuit

    backend_params: dict
        Dictionary with kwargs to define the backend. See GetBackend in module QPE.Backend for description of these kwargs
    
    method_specific_params: dict
        Dictionary which hold parameters that are not common for all algorithms

    re_initialize: bool
        Parameter that decides whether the eigenstate should be re-inintialized between different trials or not

    Methods
    --------

    get_input_states:
        input: N_states, input_states, eigen_coefs
        output: A list of lists. Every sublist describes a state vector in the computational basis.
    
    get_backend:
        inp


    Attributes
    ----------
    U
    n_digits
    method_specific params    
    re_initialise

    states:
        A list of lists, where every sublist describes the input state in the computational basis.
        This 



    """


    def __init__(self, U:Unitary, n_digits:int, N_states:int = -1, input_states: List[list] = [[]], 
                eigen_coefs: List[list] = [[]], n_shots:int = 1024, backend_params: dict = {"service":"local"}, 
                method_specific_params ={}, re_initialize:bool = True,):
        self.U = U
        self.n_digits = n_digits
        self.method_specific_params = method_specific_params
        self.re_initialize = re_initialize
        self.states, self.state_type, self.N_states = self.get_input_states(N_states, input_states, eigen_coefs)
        self.n_shots = n_shots
        self.backend_params = backend_params
        self.backend = self.get_backend()
        self.circuits = self.get_circuits()
        self.transpiled_circuits = self.get_transpiled_circuits()
        self.jobs = []
        self.job_results = []
        self.counts = []
        self.estimated_phases = []
        self.estimated_bits = []
        self.circuit_depths, self.transpiled_circuit_depths = [], []
        self.bitstring_distributions = []
        self.estimator_flag = ""


    def get_input_states(self, N_states, input_states, eigen_coefs):
        s = []
        if ((len(input_states) == 1) and (len(input_states[0]) ==0)) and (len(eigen_coefs) == 1) and (len(eigen_coefs[0]) ==0):
            if N_states > len(self.U.eigen):
                print(f"""Given value of 'N_states' is {N_states}, but Unitary has only {len(self.U.eigen)} eigenvectors.
                    Therefore 'N_states' is changed to {len(self.U.eigen)} """)
                N_states = len(self.U.eigen)
            elif N_states == -1:
                N_states = 2**self.U.dim
            for eig in self.U.eigen[:N_states]:
                s.append(eig.vector)
            state_type = "eigenvectors"
        elif len(input_states[0]) != 0:
            for state in input_states:
                if len(state) != 2**self.U.dim:
                    raise ValueError("Wrong size of input_state")
                s.append(state)
            state_type = "computational"
        elif len(eigen_coefs[0]) != 0:
            eigenvectors = self.U.eigen 
            for eigen_coef in eigen_coefs:
                if len(eigen_coef) != 2**self.U.dim:
                    raise ValueError("Wrong size of eigen_coefs")
                input_state = list(np.zeros(len(eigen_coef)))
                for i in range(len(input_state)):
                    new_eig = list(eigen_coef[i]*eigenvectors[i].vector)
                    input_state = [tot+new for tot,new in zip(input_state, new_eig)]
                norm = np.linalg.norm(input_state)
                input_state = input_state/norm
                s.append(input_state)
            state_type = "eigen combo"
        else:
            raise ValueError("No valid input to generate initial state")
        return s, state_type, len(s)

    def get_backend(self):
        backend = GetBackend(**self.backend_params)
        return backend

    def get_circuits(self):
        return [QuantumCircuit(1,1)]        

    def get_transpiled_circuits(self):
        transpiled_circuits = []
        for c in self.circuits:
            c_t = transpile(c, backend= self.backend, optimization_level=3)
            transpiled_circuits.append(c_t)
        return transpiled_circuits

    def run_circuit(self, index = -1):
        c_obj = self.transpiled_circuits[index]
        job = execute(c_obj, self.backend, shots = self.n_shots
        )
        self.jobs.append(job)
        self.wait_for_job()
        result = self.jobs[-1].result()
        self.job_results.append(result)

    def post_process(self):
        pass

    def get_depths(self):
        if len(self.circuits) == 0:
            raise ValueError("No circuits available")
        depths, transpiled_depths = [], []
        for circ, t_circ in zip(self.circuits, self.transpiled_circuits):
            d = circ.depth()
            t_d = t_circ.depth()
            depths.append(d)
            transpiled_depths.append(t_d)
        return depths, transpiled_depths        

    def get_status(self):
        if len(self.jobs) == 0:
            return "Job not submitted"
        else:
            job_status = self.jobs[-1].status()  # Query the backend server for job status.
            if job_status is JobStatus.QUEUED:
                pos = self.jobs[-1].queue_position()
                return f"Q {pos}"
            elif job_status == JobStatus.RUNNING:
                return "R"
            elif job_status == JobStatus.DONE:
                return "D"
            elif job_status == JobStatus.ERROR:
                return "E"
            elif job_status == JobStatus.VALIDATING:
                return "V"

    def wait_for_job(self, n_round = 0):
        # print("HALLAAA")
        start_time = time.time()
        max_minutes = 30
        status_print_period_minutes = 3
        print_time = time.time()
        new_status = self.get_status()
        if new_status == "D":
            waiting = False

        elif new_status == "Job not submitted":
            raise RuntimeError(f"Job not submitted")
        else:
            waiting = True
            while waiting:
                old_status = new_status
                time.sleep(2)
                new_status = self.get_status()
                current_time = time.time()
                if current_time - start_time > 60*max_minutes:
                    raise RuntimeError(f"Script has been waiting for {max_minutes} minutes\nExecution terminated")
                # print(new_status)
                if current_time - print_time > 60*status_print_period_minutes:
                    self.print_status(new_status)
                    print_time = current_time

                if new_status != old_status:
                    self.print_status(new_status)
                    
                    if new_status[0] == "D":
                        waiting = False

    def print_status(self, status):
        if status[0] == "Q":
            print(f"In queue position {status[2]}")
        elif status[0] == "R":
            print("Job is running")
        elif status[0] == "E":
            print("Error with job")
            raise RuntimeError("Error with job")
        elif status[0] == "D":
            print("Job done")
        elif status[0] == "V":
            print("Validating job")

    def draw_circuits(self, path:str = "", N_non_transpiled:int = 1,
                     N_transpiled:int = 0, show = False):
        circs_to_draw = []
        if N_non_transpiled != 0:
            if (N_non_transpiled == -1) or (N_non_transpiled > len(self.circuits)):
                N_non_transpiled = len(self.circuits)
            circs_to_draw += self.circuits[:N_non_transpiled]
        if N_transpiled != 0:
            if (N_transpiled == -1) or (N_transpiled > len(self.transpiled_circuits)):
                N_transpiled = len(self.transpiled_circuits)
            circs_to_draw += self.transpiled_circuits[:N_transpiled]

        n_circuits = len(circs_to_draw)

        if path != "" and (len(path)>4) and (path[-4:] in [".jpg", ".png"]):
            path = path[:-4]
        for i in range(n_circuits):
            s = " - Non-transpiled" if i < N_non_transpiled else " - Transpiled"
            temp_path = path + s
            drawing = circs_to_draw[i].draw("mpl")
            if show:
                drawing.show()
            if n_circuits == 1:
                drawing.savefig(temp_path + ".jpg")
            else:
                drawing.savefig(temp_path + f" - Circuit {i}.jpg")

    def get_dicts(self):
        dicts = []
        u_dict = self.U.get_dict(self.n_digits)
        backend_dict = self.get_backend_dict()
        for i in range(self.N_states):
            input_state = mat_im_to_tuple(self.states[i])
            phase_dict = self.get_phase_dict(i)
            d = {
                "Estimator": self.estimator_flag,
                "Backend": backend_dict,
                "Unitary": u_dict,
                "Bits of precision": self.n_digits,
                "Input state": input_state,
                "estimated phase": phase_dict,
                "shots": self.n_shots,
                "method_specific_params": self.method_specific_params,
                "re_initialize": self.re_initialize,
                "timestamp": datetime.now().strftime("%m_%d_%Y_%H:%M")
            }
            dicts.append(d)
        return dicts

    def get_backend_dict(self):
        d = {}
        d["service"] = self.backend_params["service"]
        d["name"] = self.backend.name()
        d["n_qubits"] = self.backend.configuration().n_qubits
        d["simulator"] = self.backend.configuration().simulator
        # d["flags"] = self.backend_params
        return d

    def get_phase_dict(self, s_i):
        d = {}
        d["phi"] = self.estimated_phases[s_i]
        d["bitstring"] = self.estimated_bits[s_i]
        if self.estimator_flag in ["Kitaev", "Iterative"]:
            # print(self.e)
            d["distribution"] = {self.estimated_bits[s_i]: 1}
        elif self.estimator_flag == "QFT":
            d["distribution"] = self.bitstring_distributions[s_i]
        return d

    def get_bitstrings(self):
        # Return a list of the estimated bitstring orresponding to each state
        bitstrings = []
        if self.estimator_flag in ["Kitaev", "Iterative"]:
            for i in range(len(self.states)):
                b = "".join([str(int(bit)) for bit in self.estimated_bits[i,:].copy().flatten()])
                bitstrings.append(b)
        elif self.estimator_flag == "QFT":
            return self.estimated_bits
            #FIX THIS
        if len(bitstrings) != self.N_states:
            raise ValueError("Something is WRONG")
        return bitstrings
    
    def get_bitstring_distribution(self, s_i):
        if self.estimator_flag in ["Kitaev", "Iterative"]:
            return {
                self.bitstrings[s_i]: 1
            }
        elif self.estimator_flag == "QFT":
            return self.bitstring_distributions[s_i]
    
    def dump_to_json(self, path: Union[List[str], str]):
        dicts = self.get_dicts()
        n_dicts = len(dicts)
        if isinstance(path, str):
            if path[-5:] == ".json":
                path = path[:-5]
            paths = [f"{path}_state{s_i}.json" for s_i in range(n_dicts)]
        elif isinstance(path, list):
            paths = []
            for p in path:
                if (len(p) < 5) or (p[-5:] != ".json"):
                    p += ".json"
                paths.append(p)
        
        for d, p2 in zip(dicts, paths):
            with open(p2, "w") as write_file:
                json.dump(d, write_file, indent = True)
            print(f"Data saved to directory {p2}")

if __name__ == "__main__":
    pass


from math import pi
from qiskit import QuantumCircuit, transpile, execute
import numpy as np
from QPE.HelpFunctions import bin_to_float, count_conversion
from QPE.PhaseEstimator import PhaseEstimator
from QPE.Unitary import Unitary
import time
from typing import List

def IterativeCircuit(U, n_digits, n_round, measurements, states, re_initialize):
    dim = U.dim + 1
    eigen = U.eigen
    N_states = len(states)
    circ = QuantumCircuit(dim,N_states)

    if measurements.shape[0] != N_states or measurements.shape[1] != n_digits:
        raise ValueError("Wrong dimensions of array 'measurements'")

    circ.initialize([1,0], 0)
    target_qubits = list(range(1, dim))
    for s_i, state in enumerate(states):

        circ.initialize(state, target_qubits)
        circ.h(0)
        previous_bits = list(np.flip(measurements[s_i,:n_round].copy().flatten()))
        previous_bit_string = [str(int(bit)) for bit in previous_bits]
        # print(previous_bit_string)
        phi_shift = - bin_to_float(previous_bit_string)/2
        angle_shift = 2*pi * phi_shift
        # for n, outcome in enumerate(previous_bits):
        #     angle += (-2*pi* 2**(- 2 - n))*outcome
        circ.rz(angle_shift,0)
        power = 2**(n_digits-1-n_round)
        # print(power)
        u = U.get_gate(power, 1)
        circ.append(u, list(range(dim)))
        circ.h(0)
        circ.measure(0, s_i)
        if s_i < N_states - 1 :
            circ.barrier()
            circ.reset(list(range(dim)))
    return circ
    
            
class Iterative(PhaseEstimator):
    def __init__(self, U:Unitary, n_digits:int, N_states:int = -1, input_states: List[list] = [[]], 
                eigen_coefs: List[list] = [[]], n_shots:int = 1024, backend_params: dict = {"service":"local"}, 
                method_specific_params ={}, re_initialize:bool = True,):

        super().__init__(U=U, n_digits=n_digits, N_states=N_states, re_initialize=re_initialize,
                n_shots=n_shots, backend_params=backend_params, input_states=input_states,
                eigen_coefs=eigen_coefs, method_specific_params=method_specific_params)

        self.circuits = []
        self.estimator_flag = "Iterative"
        
    def get_one_circuit(self, n_round = 0, measurements = None):
        if not isinstance(measurements, type(np.zeros(0))):
            circ = QuantumCircuit(1,1)
        else:
            circ = IterativeCircuit(self.U, self.n_digits, 
                                n_round, measurements, self.states, self.re_initialize)
        return circ

    def run_circuit(self):
        measurements = np.zeros((self.N_states, self.n_digits))
        for n_round in range(self.n_digits):
            circ = self.get_one_circuit(n_round, measurements)
            self.circuits.append(circ)
            transpiled_circuit = transpile(circ, backend=self.backend)
            self.transpiled_circuits.append(transpiled_circuit)
            self.jobs.append(execute(transpiled_circuit, self.backend, shots=self.n_shots))
            time.sleep(1) #We need to wait a bit, otherwise get_status=None
            self.wait_for_job(n_round=n_round)
            result = self.jobs[-1].result()
            self.job_results.append(result)
            counts = dict(result.get_counts())

            ones = count_conversion(counts)

            majority_bit_result = [1 if n_ones > self.n_shots/2 else 0 for n_ones in reversed(ones)]
            
            # print(majority_bit_result)
            measurements[:,n_round] = majority_bit_result
            # print(measurements)
        self.measurements = measurements
    
    def post_process(self):
        # estimated_bits = np.zeros((self.N_states, self.n_digits))
        for i in range(self.N_states):
            phase = 0
            bits = np.flip(self.measurements[i,:].copy())
            # estimated_bits[i,:] = bits
            bit_list = list(bits.copy().flatten())
            bitstring = "".join([str(int(bit)) for bit in bit_list])
            phase = bin_to_float(bitstring)
            # for b_i, b in enumerate(bits):
            #     phase += 2**(- 1 - b_i)*b
            self.estimated_phases.append(phase)
            self.estimated_bits.append(bitstring)




import json
from QPE.HelpFunctions import mat_im_to_tuple, mat_tuple_to_im
from QPE.Unitary import Unitary
from QPE.QFT import QFT
from QPE.Kitaev import Kitaev
from QPE.Iterative import Iterative

def unpack_PE_from_json(path):

    with open(path, "r") as json_object:
        d = json.load(json_object)

    estimator = d["Estimator"]
    
    tuple_matrix = d["Unitary"]["data"]
    print(tuple_matrix)
    print(type(tuple_matrix))
    im_matrix = mat_tuple_to_im(tuple_matrix)
    U = Unitary(im_matrix)

    n_digits = d["Bits of precision"]
    input_states = [mat_tuple_to_im(d["Input state"])]
    n_shots = d["shots"]
    backend_params = {"service": d["Backend"]["service"]}
    if backend_params["service"] in ["IBMQ", "QI"]:
        backend_params["backend_name"] = d["Backend"]["name"]
    method_specific_params = d["method_specific_params"]
    re_initialize = d["re_initialize"]
    
    if estimator == "QFT":
        PE = QFT
    elif estimator == "Kitaev":
        PE = Kitaev
    elif estimator == "Iterative":
        PE = Iterative
    PE_object = PE(U, n_digits, n_shots=n_shots,
            backend_params=backend_params, input_states=input_states,
            method_specific_params=method_specific_params, re_initialize=re_initialize)
    

    PE_object.estimated_phases = [d["estimated phase"]["phi"]]
    PE_object.estimated_bits = [d["estimated phase"]["bitstring"]]
    PE_object.bitstring_distributions = [d["estimated phase"]["distribution"]]
    
    return PE_object

if __name__ == "__main__":
    p = "example_experiment_hardware.json"
    PE = unpack_PE_from_json(p)


from QPE.PhaseEstimator import PhaseEstimator
from QPE.Unitary import Unitary
from qiskit import QuantumCircuit
from math import pi
import numpy as np
from QPE.HelpFunctions import count_conversion, float_to_bin, get_sub_bitstring_counter, bin_to_float
from typing import List
from qiskit.visualization import plot_histogram

def QFTCircuit(U, n_digits, input_states:List[list], n_rotations = 2, re_initialize = True):

    dim = U.dim + n_digits
    N_states = len(input_states)
    circ = QuantumCircuit(dim,(dim-U.dim)*N_states)

    if (n_rotations > n_digits - 1) or n_rotations == -1:
        n_rotations = n_digits -1

    for s_i, state in enumerate(input_states):
        circ.barrier()
        for i in range (n_digits):
            circ.initialize([1,0],i)
      
        target_qubits = list(range(n_digits, dim))
        circ.initialize(state, target_qubits)
        circ.barrier()
        circ.h(list(range(n_digits)))

        for i in range(n_digits):
            power = 2**i
            controller = n_digits -1 - i
            u = U.get_gate(power, ctrl = 1)
            circ.append(u, [controller]+list(range(n_digits, dim)))

        for i in range(0, n_digits):
            circ.h(i)
            for j in range(i+1, i + n_rotations + 1):
                if j<(n_digits):
                    angle = -pi*(2**(i-j))
                    circ.crz(angle, i, j)

        for i in range (n_digits):
            circ.barrier()
            circ.measure(i,i+s_i*n_digits)
    return circ

class QFT(PhaseEstimator):
    def __init__(self, U:Unitary, n_digits:int, N_states:int = -1, input_states: List[list] = [[]], 
                eigen_coefs: List[list] = [[]], n_shots:int = 1024, backend_params: dict = {"service":"local"}, 
                method_specific_params ={"n_rotations":-1}, re_initialize:bool = True,):
        
        super().__init__(U=U, n_digits=n_digits, N_states=N_states, re_initialize=re_initialize,
                n_shots=n_shots, backend_params=backend_params, input_states=input_states,
                eigen_coefs=eigen_coefs, method_specific_params=method_specific_params)
        
        self.estimator_flag = "QFT"

    def get_circuits(self):
        print(self.method_specific_params)
        circ = QFTCircuit(self.U, self.n_digits, self.states, self.method_specific_params["n_rotations"])
        return [circ]
        # Return the required QFT Circuit

    def post_process(self):
        self.wait_for_job()
        self.bitstring_distributions = []
        #for r_i in range(len(self.job_results)):
        counts = dict(self.job_results[-1].get_counts())
        b_dict = get_sub_bitstring_counter(counts, self.N_states)
        b_dict.reverse()
        #print(b_dict)
        self.bitstring_distributions = b_dict
        estimated_bits = []
        for i in range(self.N_states):
            d = self.bitstring_distributions[i]
            maximum_counts = 0
            for key in d:
                if d[key] > maximum_counts:
                    maximum_counts = d[key]
                    bitstring = key
            estimated_bits.append(bitstring)
        
        estimated_bits.reverse()
        self.estimated_bits = estimated_bits
        self.estimated_phases = [bin_to_float(s) for s in self.estimated_bits]
    
    def plot_histograms(self, s_i):
        d = self.bitstring_distributions[s_i]
        # title = fr"Input state $\varphi_{s_i}$"
        x = [key for key in d]
        n = [d[key] for key in d]
        hist = plot_histogram(d)
        return hist
        # print(n)
        # plt.hist(d)
        # plt.show()
        # fig.show()
        # return fig



from math import pi
from qiskit import QuantumCircuit
import numpy as np
from QPE.HelpFunctions import bin_to_float, count_conversion, calculate_bitstring_distribution
from QPE.Unitary import Unitary
from QPE.PhaseEstimator import PhaseEstimator

def SuperIterativeCircuit(U, n_digits, states):
    dim = U.dim + 1
    N_states = len(states)
    n_measurents = (2**n_digits - 1) * N_states
    circ = QuantumCircuit(dim, n_measurents)
    target_qubits = list(range(1, dim))
    
    classical_index = 0
    for s_i, state in enumerate(states):
        for k in range(n_digits):
            power = 2**(n_digits - 1 - k)
            n_shifts = 2**k
            u = U.get_gate(power, 1)
            for shift in range(n_shifts):
                circ.initialize(state, target_qubits)
                circ.initialize([1,0], 0)
                circ.barrier()
                circ.h(0)
                circ.append(u, list(range(dim)))
                angle_shift = shift*pi/n_shifts
                circ.rz(angle_shift,0)
                circ.h(0)
                circ.measure(0, classical_index)
                classical_index += 1
                if classical_index < n_measurents:
                    circ.barrier()
                    circ.reset(list(range(dim)))
    
    return circ


class SuperIterative(PhaseEstimator):
    def __init__(self, U: Unitary, n_digits:int, N_states = 1, n_shots = 1024, backend_params: dict = {}, input_states = [[]], eigen_coefs = [[]]):
        super().__init__(U, n_digits, N_states, n_shots, backend_params, input_states, eigen_coefs)
        self.estimator_flag = "SuperIterative"
        
    def get_circuits(self):
        circ = [SuperIterativeCircuit(self.U, self.n_digits, self.states)]
        circ[0].draw("mpl").savefig("hallooo.jpg")
        return circ
    
    def post_process(self):
        counts = dict(self.job_results[0].get_counts())
        ones = count_conversion(counts)
        m_per_state = 2**self.n_digits - 1
        ones_per_state = [list(ones[m_per_state*i : m_per_state*(i+1)]) for i in range(self.N_states)]
        # new_order_ones_per_state = []
        for state_ones in ones_per_state:
            right_order_ones = state_ones[::-1]
            d = calculate_bitstring_distribution(right_order_ones, self.n_digits, self.n_shots)
            self.bitstring_distributions.append(d)
        estimated_bits = []
        for dist in self.bitstring_distributions:
            max_p = 0
            for key in dist:
                if dist[key] > max_p:
                    max_bitstring = key
                    max_p = dist[key]
            estimated_bits.append(max_bitstring)
        self.estimated_bits = estimated_bits
        self.estimated_phases = [bin_to_float(s) for s in self.estimated_bits]

if __name__ == "__main__":
    U = Unitary(random=True, random_state=123)
    eigen1 = [list(U.eigen[0].vector.copy())]
    print(eigen1)
    


    c = SuperIterativeCircuit(U, 4, eigen1)
    f = c.draw("mpl")
    f.savefig("SuperIterative.jpg")


import json
from typing import Set
from qiskit import IBMQ, Aer
from qiskit.providers.ibmq import least_busy
from quantuminspire.credentials import enable_account as load_QI
from quantuminspire.qiskit import QI

def Login(service="IBMQ", token_path = "API_tokens.json", token = ""):


    if service not in ["IBMQ", "QI"]:
        raise KeyError('Choose service "IBMQ" or "QI"')

    if token:
        API_key = token
    else:
        try:
            with open(token_path, "r") as f:
                token_dict = json.load(f)
        except:
            raise FileNotFoundError(f"No json file {token_path} found")
        try:
            API_key = token_dict[service]
        except:
            raise KeyError(f"No token in {token_path} for service {service}")

    if service == "IBMQ":
        try:
            IBMQ.load_account()
        except:
            API_key = token_dict["IBMQ"]
            IBMQ.enable_account(API_key)
    elif service == "QI":
        API_key = token_dict["QI"]
        load_QI(API_key)
        QI.set_authentication()

def GetBackend(service="IBMQ", backend_name = "", flags:list = [], token_path = "API_tokens.json", token = ""):

    if service not in ["IBMQ", "QI", "local"]:
        raise KeyError('Choose service "IBMQ","QI" or "local"')

    if service in ["IBMQ", "QI"]:
        Login(service, token_path, token)
    
    sim = lambda x: x.configuration().simulator,
    Q5 = lambda x: x.configuration().n_qubits == 5 and not x.configuration().simulator

    if service == "IBMQ":
        provider = IBMQ.get_provider()
        if backend_name == "":
            if "5qubit" in flags:
                backend = least_busy(provider.backends(
                    filters= Q5
                    ))
            elif "simulator" in flags:
                backend = least_busy(provider.backends(
                    filters= sim
                    ))
        else:
            backend = provider.get_backend(backend_name)
    
    elif service == "QI":
        backend = QI.get_backend("Starmon-5")
        print(backend)
    
    elif service == "local":
        backend = Aer.get_backend("aer_simulator")

    print(service)
    return backend

filters = {
    "simulator": lambda x: x.configuration().simulator,
    "5qubit": lambda x: x.configuration().n_qubits == 5
                                   and not x.configuration().simulator
 }

if __name__ == "__main__":
    Login("IBMQ")
    #print(IBMQ.providers())
    backend = GetBackend("IBMQ")
    


from re import U
from QPE.PhaseEstimator import PhaseEstimator
from QPE.Unitary import Unitary
from qiskit import QuantumCircuit
from math import pi
import numpy as np
from QPE.HelpFunctions import count_conversion, float_to_bin, bin_to_float
from typing import List

def KitaevCircuit(U, n_digits, states, re_initialize = True):
    dim = U.dim + 1
    N_states = len(states)
    N_class_regs = 2*N_states*(n_digits-2)
    circ = QuantumCircuit(dim, N_class_regs)
    circ.barrier()
    # circ.initialize([1,0], 0)
    target_qubits = list(range(1, dim))
    n_round = 0
    # class_reg = 0
    for state in states:
        for n in range(n_digits-2):
            for rot in range(2):
                circ.initialize([1,0], 0)
                if re_initialize or (n_round == 0):
                    circ.initialize(state, target_qubits)
                circ.barrier()
                circ.h(0)
                if rot == 1:
                    circ.rz(pi/2,0)
                power = 2**(n_digits - 3 - n)
                u = U.get_gate(power, 1)
                circ.append(u, list(range(dim)))
                circ.h(0)
                
                circ.measure(0,n_round)
                n_round += 1
                if n_round < N_class_regs: 
                    circ.barrier()

    return circ

class Kitaev(PhaseEstimator):
    def __init__(self, U:Unitary, n_digits:int, N_states:int = -1, input_states: List[list] = [[]], 
                eigen_coefs: List[list] = [[]], n_shots:int = 1024, backend_params: dict = {"service":"local"}, 
                method_specific_params ={}, re_initialize:bool = True,):

        if n_digits < 3:
            raise ValueError(f"Kitaevs algorithm estimates at least 3 digits, but n_digits={n_digits} was given")
        super().__init__(U=U, n_digits=n_digits, N_states=N_states, re_initialize=re_initialize,
                n_shots=n_shots, backend_params=backend_params, input_states=input_states,
                eigen_coefs=eigen_coefs, method_specific_params=method_specific_params)
        self.estimator_flag = "Kitaev"

    def get_circuits(self):
        circ = KitaevCircuit(self.U, self.n_digits, self.states)
        return [circ]

    def post_process(self):
        self.wait_for_job()
        result = self.jobs[-1].result()
        self.job_results.append(result)
        counts = np.flip(count_conversion(dict(result.get_counts())))

        ones = np.array(counts).reshape((self.N_states, self.n_digits-2, 2))

        rho = np.zeros((self.N_states, self.n_digits-2))
        for (N, n), _ in np.ndenumerate(rho):
            n1_cos = ones[N,n,0]
            n0_cos = self.n_shots - n1_cos
            cos = (n0_cos - n1_cos)/self.n_shots
            n1_sin = ones[N,n,1]
            n0_sin = self.n_shots - n1_sin
            sin = (n1_sin - n0_sin)/self.n_shots
            if cos == 0:
                if abs(sin-1) < 0.9:
                    angle = pi/2
                else:
                    angle = 3*pi/2
            else:
                angle = np.arctan(sin/cos)
                if cos < 0:
                    angle += pi
            if angle < 0:
                angle += 2*pi
            rho_Nn = angle/(2*pi)
            rho[N,n] = rho_Nn

        alpha = np.zeros((self.N_states, self.n_digits))

        for N in range(self.N_states):
            least_significant_rho = rho[N,0]
            #print(least_significant_rho)
            alpha_bits = float_to_bin(least_significant_rho, 3)
            # print(alpha_bits)
            alpha_bit_list = [int(b) for b in alpha_bits][::-1]
            # print(alpha_bit_list)
            alpha[N,0:3] = alpha_bit_list
            # print(N, alpha_bit_list)

        for N in range(self.N_states):
            for j in range(1, self.n_digits-2):
                rho_j = rho[N,j]
                # print(N,j,rho_j)
                alpha_eight = alpha[N,j]
                alpha_quarter = alpha[N,j+1]
                diff = rho_j - (alpha_quarter*1/4 + alpha_eight*1/8)
                if abs(diff) > 0.5:
                    diff = 1 - abs(diff)
                if diff <= 0.25:
                    alpha_half = 0
                else:
                    alpha_half = 1
                alpha[N, j+2] = alpha_half
        
        estimated_phases = []
        for N in range(self.N_states):
            bits = list(np.flip(alpha[N,:].flatten()))
            bitstring = "".join([str(int(bit)) for bit in bits])
            # print(bitstring)
            phase = bin_to_float(bitstring)
            estimated_phases.append(phase)
            
        self.estimated_phases = estimated_phases
        # estimated_bits = []
        bit_array = np.flip(alpha, axis = 1)
        for i in range(self.N_states):
            bit_list = [str(int(bit)) for bit in bit_array[i,:].copy()]
            bitstring = "".join(bit_list)
            self.estimated_bits.append(bitstring)

if __name__ == "__main__":
    from input_experiments import U2, U1
    from qiskit.visualization import plot_histogram
    u = Unitary(random= True, dim = 1)
    k = Kitaev(U2, 4, N_states=2, n_shots = 60)
    # u = Unitary(random= True, dim = 1)

    bit_dict0, bit_dict1 = {},{}

    # for i in range(60):
        
    k = Kitaev(U1, 4, N_states=2, n_shots = 60)
    k.run_circuit()
    k.post_process()
        # print(k.estimated_bits)
        # b0, b1 = k.estimated_bits
        # if b0 in bit_dict0:
        #     bit_dict0[b0] += 1
        # else:
        #     bit_dict0[b0] = 1
        # if b1 in bit_dict1:
        #     bit_dict1[b1] += 1
        # else:
        #     bit_dict1[b1] = 1
        
        # if i % 10 == 0:
        #     print(f"{i/10} percent done")

    # h0 = plot_histogram(bit_dict0, title="State 0")
    # h1 = plot_histogram(bit_dict1, title= "State 1")

    # h0.savefig("Kitaev_0_low_shots_60.png")
    # h1.savefig("Kitaev_1_low_shots_60.png")




    print(k.estimated_bits)
    print(k.U.get_phi_bitstrings(5))


import numpy as np
from QPE.HelpFunctions import float_to_bin, mat_im_to_tuple
from qiskit import QuantumCircuit
from qiskit.quantum_info.operators import Operator
from scipy.stats import unitary_group
from math import pi

# np.array()

class Unitary():
    def __init__(self, operator = [], random = False, random_state=None, dim = 1, name = ""):
        self.name = name
        self.operator = self.get_operator(operator, random, random_state, dim)
        self.dim = self.get_dim()
        self.eigen = self.get_eigen()
        self.vectors = self.get_vectors()
        self.phis  = self.get_phis()
    
    def get_operator(self, operator, random, random_state, dim):

        if random == True:
            self.dim = dim
            operator = self.gen_random(random_state)
        else:
            try:
                operator = Operator(operator)
            except:
                raise(ValueError('insert a list, an operator or a circuit'))
        return operator

    def gen_random(self, random_state):
        return Operator(unitary_group.rvs(2**self.dim, random_state=random_state))

    def get_dim(self):
        return int(np.log2(self.operator.dim[0]))

    def get_gate(self, power=1, ctrl=0):  

        changed_op = self.operator.power(power)

        circ_ctr = QuantumCircuit(self.dim)
        circ_ctr.append(changed_op,list(range(self.dim)))
        lab = r'$U^{'+str(power)+r'}$' if power != 1 else 'U'

        if ctrl != 0:
            changed_op = circ_ctr.to_gate(label = lab).control(ctrl)
        else:
            changed_op = circ_ctr.to_gate(label = lab)
    
        return changed_op
        
        # Return gate that can be appended to circuit U^power with ctrl control qubits
        # ex.
        # circ.append(U, [0,1,2])
        # if ctrl = 1 then the control qubit is the fist one
        # if ctrl = 2 then the control qubits is the fist two
    
    def get_eigen(self):
        # self.operator.data
        eigval, eigvect = np.linalg.eig(self.operator.data)
        # print(eigval, eigvect)
        eig = np.concatenate((np.array(eigval).reshape(len(eigval),1), eigvect), axis=1)
        eigenvectors = []
        for i in range(len(eigval)):
            value = eig[i,0]
            vector = eig[i,1:]
            el = Eigenvector(vector, value, i)
            eigenvectors.append(el)

        return eigenvectors
        # Return a matrix 
        # Each raw is [eigenvalue, eigenvector[0], eigenvector[1],...eigenvector[2**dim]]
    
    def get_vectors(self):
        vectors = []
        for e in self.eigen:
            vectors.append(e.vector)
        return vectors

    def get_phis(self):
        phis = []
        for e in self.eigen:
            phis.append(e.phi)
        return phis
    
    def get_phi_bitstrings(self, n_digits):
        b_list = []
        for phi in self.phis:
            b = float_to_bin(phi, n_digits)
            b_list.append(b)
        return b_list
    
    def get_dict(self, n_digits):
        d = {}
        d["dimension"] = self.dim
        d["data"] = mat_im_to_tuple(self.operator.data)
        d["eigenvectors"] = mat_im_to_tuple(self.vectors)
        d["phis"] = mat_im_to_tuple(self.phis)
        d["phi_bitstrings"] = self.get_phi_bitstrings(n_digits)
        return d


class Eigenvector():
    def __init__(self, vector, value, index):
        self.vector = vector
        self.value = value 
        self.index = index
        self.phi = self.get_phi()    
    
    def get_phi(self):
        phi = np.angle(self.value)/(2*pi)
        if phi < 0:
            phi += 1
        return phi
        #Return the phase , i.e. a number between 0 and 1
        # The numpy.angle function could probably be useful
    
    def get_phi_bitstring(self, n_digits):
        bitstring = float_to_bin(self.phi, n_digits)
        return bitstring


if __name__ == "__main__":
    u = Unitary(random=True)
    #print(u.operator.data)
    d = u.get_dict(3)
    print(d)
    # k = u.eigen[0].get_phi_bitstring(n_digits = 10)
    # print(k)
    # phi = u.eigen[0].get_phi()
    # print(phi)
